@@ -16,11 +16,7 @@ EnumValue = Union[str, int]
 EnumType = Dict[str, EnumValue]
 
 functions = {}  # type: Dict[str, FunctionType]
-lxc_functions = {}  # type: Dict[str, FunctionType]
-qemu_functions = {}  # type: Dict[str, FunctionType]
 enums = defaultdict(dict)  # type: Dict[str, EnumType] # { enumType: { enumConstant: enumValue } }
-lxc_enums = defaultdict(dict)  # type: Dict[str, EnumType] # { enumType: { enumConstant: enumValue } }
-qemu_enums = defaultdict(dict)  # type: Dict[str, EnumType] # { enumType: { enumConstant: enumValue } }
 event_ids = []  # type: List[str]
 params = []  # type: List[Tuple[str, str]] # [ (paramName, paramValue)... ]
 
@@ -102,12 +98,12 @@ class docParser(xml.sax.handler.ContentHandler):
                 self.function_return_field = attrs.get('field', '')
         elif tag == 'enum':
             # enums come from header files, hence virterror.h
-            if attrs['file'] in libvirt_headers + ["virerror", "virterror"]:
+            files = libvirt_headers + ["virerror",
+                                       "virterror",
+                                       "libvirt-lxc",
+                                       "libvirt-qemu"]
+            if attrs['file'] in files:
                 enum(attrs['type'], attrs['name'], attrs['value'])
-            elif attrs['file'] == "libvirt-lxc":
-                lxc_enum(attrs['type'], attrs['name'], attrs['value'])
-            elif attrs['file'] == "libvirt-qemu":
-                qemu_enum(attrs['type'], attrs['name'], attrs['value'])
         elif tag == "macro":
             if "string" in attrs:
                 params.append((attrs['name'], attrs['string']))
@@ -119,37 +115,21 @@ class docParser(xml.sax.handler.ContentHandler):
             # functions come from source files, hence 'virerror.c'
             if self.function:
                 assert self.function_return
-                if self.function_module in libvirt_headers + \
-                        ["event", "virevent", "virerror", "virterror"]:
+
+                modules = libvirt_headers + ["event",
+                                             "virevent",
+                                             "virerror",
+                                             "virterror",
+                                             "libvirt-lxc",
+                                             "libvirt-qemu"]
+                files = ["python", "python-lxc", "python-qemu"]
+
+                if (self.function_module in modules or
+                    self.function_file in files):
                     function(self.function, self.function_descr,
                              self.function_return, self.function_args,
                              self.function_file, self.function_module,
                              self.function_cond)
-                elif self.function_module == "libvirt-lxc":
-                    lxc_function(self.function, self.function_descr,
-                                 self.function_return, self.function_args,
-                                 self.function_file, self.function_module,
-                                 self.function_cond)
-                elif self.function_module == "libvirt-qemu":
-                    qemu_function(self.function, self.function_descr,
-                                  self.function_return, self.function_args,
-                                  self.function_file, self.function_module,
-                                  self.function_cond)
-                elif self.function_file == "python":
-                    function(self.function, self.function_descr,
-                             self.function_return, self.function_args,
-                             self.function_file, self.function_module,
-                             self.function_cond)
-                elif self.function_file == "python-lxc":
-                    lxc_function(self.function, self.function_descr,
-                                 self.function_return, self.function_args,
-                                 self.function_file, self.function_module,
-                                 self.function_cond)
-                elif self.function_file == "python-qemu":
-                    qemu_function(self.function, self.function_descr,
-                                  self.function_return, self.function_args,
-                                  self.function_file, self.function_module,
-                                  self.function_cond)
                 self.in_function = False
         elif tag == 'arg':
             if self.in_function:
@@ -179,18 +159,6 @@ def function(name: str, desc: str, ret: ArgumentType, args: List[ArgumentType],
     functions[name] = (desc, ret, args, file, module, cond)
 
 
-def qemu_function(name: str, desc: str, ret: ArgumentType, args: List[ArgumentType], file: str, module: str, cond: str) -> None:
-    if onlyOverrides and name not in qemu_functions:
-        return
-    qemu_functions[name] = (desc, ret, args, file, module, cond)
-
-
-def lxc_function(name: str, desc: str, ret: ArgumentType, args: List[ArgumentType], file: str, module: str, cond: str) -> None:
-    if onlyOverrides and name not in lxc_functions:
-        return
-    lxc_functions[name] = (desc, ret, args, file, module, cond)
-
-
 def enum(type: str, name: str, value: EnumValue) -> None:
     if (name.startswith('VIR_DOMAIN_EVENT_ID_') or
             name.startswith('VIR_NETWORK_EVENT_ID_')):
@@ -213,27 +181,16 @@ def enum(type: str, name: str, value: EnumValue) -> None:
         value = 1
     elif value == 'VIR_DOMAIN_AFFECT_CONFIG':
         value = 2
-    if onlyOverrides and name not in enums[type]:
-        return
-    enums[type][name] = value
-
-
-def lxc_enum(type: str, name: str, value: EnumValue) -> None:
-    if onlyOverrides and name not in lxc_enums[type]:
-        return
-    lxc_enums[type][name] = value
-
-
-def qemu_enum(type: str, name: str, value: EnumValue) -> None:
-    if value == 'VIR_DOMAIN_AGENT_RESPONSE_TIMEOUT_BLOCK':
+    elif value == 'VIR_DOMAIN_AGENT_RESPONSE_TIMEOUT_BLOCK':
         value = -2
     elif value == 'VIR_DOMAIN_AGENT_RESPONSE_TIMEOUT_DEFAULT':
         value = -1
     elif value == 'VIR_DOMAIN_AGENT_RESPONSE_TIMEOUT_NOWAIT':
         value = 0
-    if onlyOverrides and name not in qemu_enums[type]:
+
+    if onlyOverrides and name not in enums[type]:
         return
-    qemu_enums[type][name] = value
+    enums[type][name] = value
 
 
 #######################################################################
@@ -243,12 +200,6 @@ def qemu_enum(type: str, name: str, value: EnumValue) -> None:
 #
 #######################################################################
 
-functions_skipped = {
-    "virConnectListDomains",
-}
-lxc_functions_skipped = set()  # type: Set[str]
-qemu_functions_skipped = set()  # type: Set[str]
-
 skipped_types = {
     # 'int *': "usually a return type",
     'virConnectDomainEventCallback': "No function types in python",
@@ -343,8 +294,6 @@ py_types = {
 }  # type: Dict[str, Tuple[str, str, str, str]]
 
 
-unknown_types = defaultdict(list)  # type: Dict[str, List[str]]
-
 #######################################################################
 #
 #  This part writes the C <-> Python stubs libvirt.[ch] and
@@ -482,13 +431,9 @@ skip_impl = {
     'virDomainAuthorizedSSHKeysSet',
     'virDomainGetMessages',
     'virNodeDeviceGetAutostart',
-}
 
-lxc_skip_impl = {
     'virDomainLxcOpenNamespace',
-}
 
-qemu_skip_impl = {
     'virDomainQemuMonitorCommand',
     'virDomainQemuAgentCommand',
 }
@@ -616,13 +561,10 @@ skip_function = {
     'virDomainFSInfoFree',  # only useful in C, python code uses list
     'virDomainIOThreadInfoFree',  # only useful in C, python code uses list
     'virDomainInterfaceFree',  # only useful in C, python code uses list
-}
 
-lxc_skip_function = {
     "virDomainLxcEnterNamespace",
     "virDomainLxcEnterSecurityLabel",
-}
-qemu_skip_function = {
+
     # "virDomainQemuAttach",
     'virConnectDomainQemuMonitorEventRegister',  # overridden in -qemu.py
     'virConnectDomainQemuMonitorEventDeregister',  # overridden in -qemu.py
@@ -635,39 +577,115 @@ function_skip_python_impl = {
                       # be exposed in bindings
 }
 
-lxc_function_skip_python_impl = set()  # type: Set[str]
-qemu_function_skip_python_impl = set()  # type: Set[str]
-
 function_skip_index_one = {
     "virDomainRevertToSnapshot",
 }
 
 
-def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[str], include: IO[str]) -> int:
-    """
-    :returns: -1 on failure, 0 on skip, 1 on success.
-    """
-    try:
-        if module == "libvirt":
-            (desc, ret, args, file, mod, cond) = functions[name]
-            skip_function2, skip_impl2 = skip_function, skip_impl
-        elif module == "libvirt-lxc":
-            (desc, ret, args, file, mod, cond) = lxc_functions[name]
-            skip_function2, skip_impl2 = lxc_skip_function, lxc_skip_impl
-        elif module == "libvirt-qemu":
-            (desc, ret, args, file, mod, cond) = qemu_functions[name]
-            skip_function2, skip_impl2 = qemu_skip_function, qemu_skip_impl
-        else:
-            raise ValueError(module)
-    except Exception:
-        print("failed to get function %s infos" % name)
+def validate_function(name):
+    (desc, ret, args, file, mod, cond) = functions[name]
+
+    if name in skip_function:
+        return []
+    if name in skip_impl:
+        return []
+
+    failed = False
+    unknown = []
+    for a_name, a_type, a_info in args:
+        # This should be correct
+        if a_type[0:6] == "const ":
+            a_type = a_type[6:]
+
+        if a_type in skipped_types:
+            return []
+
+        if a_type not in py_types:
+            unknown.append(a_type)
+
+    r_type, r_info, r_field = ret
+    if r_type in skipped_types:
+        return []
+    if r_type != 'void' and r_type not in py_types:
+        unknown.append(r_type)
+
+    return unknown
+
+
+def validate_functions():
+    unknown_types = defaultdict(list)  # type: Dict[str, List[str]]
+    funcs_failed = []  # type: List[str]
+
+    for name in sorted(functions):
+        unknown = validate_function(name)
+        if unknown:
+            funcs_failed.append(name)
+            for thetype in unknown:
+                unknown_types[thetype].append(name)
+
+    if unknown_types:
+        print("Missing type converters: ")
+        for type, count in unknown_types.items():
+            print("%s:%d " % (type, len(count)))
+
+    for f in funcs_failed:
+        print("ERROR: failed %s" % f)
+
+    if funcs_failed:
         return -1
 
-    if name in skip_function2:
-        return 0
-    if name in skip_impl2:
-        # Don't delete the function entry in the caller.
-        return 1
+    return 0
+
+
+def skip_both_impl(name: str) -> bool:
+    if name in skip_function:
+        return True
+
+    (desc, ret, args, file, mod, cond) = functions[name]
+
+    for a_name, a_type, a_info in args:
+        # This should be correct
+        if a_type[0:6] == "const ":
+            a_type = a_type[6:]
+
+        if a_type in skipped_types:
+            return True
+
+    r_type, r_info, r_field = ret
+    if r_type in skipped_types:
+        return True
+
+    return False
+
+
+def skip_c_impl(name: str) -> bool:
+    if skip_both_impl(name):
+        return True
+
+    if name in skip_impl:
+        return True
+
+    return False
+
+
+def skip_py_impl(name: str) -> bool:
+    if skip_both_impl(name):
+        return True
+
+    if name in function_skip_python_impl:
+        return True
+
+    return False
+
+
+def print_function_wrapper(package: str, name: str, output: IO[str], export: IO[str], include: IO[str]) -> bool:
+    """
+    :returns: True if generated, False if skipped
+    """
+    (desc, ret, args, file, mod, cond) = functions[name]
+
+    if skip_c_impl(name):
+        return False
 
     c_call = ""
     format = ""
@@ -701,10 +719,7 @@ def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[s
                 c_call += ", "
             c_call += "%s" % (a_name)
         else:
-            if a_type in skipped_types:
-                return 0
-            unknown_types[a_type].append(name)
-            return -1
+            raise Exception("Unexpected type %s in function %s" % (a_type, name))
     if format:
         format += ":%s" % (name)
 
@@ -734,10 +749,7 @@ def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[s
             ret_convert += "    free(c_retval);\n"
         ret_convert += "    return py_retval;\n"
     else:
-        if r_type in skipped_types:
-            return 0
-        unknown_types[r_type].append(name)
-        return -1
+        raise Exception("Unexpected type %s in function %s" % (r_type, name))
 
     if cond:
         include.write("#if %s\n" % cond)
@@ -745,18 +757,9 @@ def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[s
         output.write("#if %s\n" % cond)
 
     include.write("PyObject * ")
-    if module == "libvirt":
-        include.write("libvirt_%s(PyObject *self, PyObject *args);\n" % (name))
-        export.write("    { (char *)\"%s\", libvirt_%s, METH_VARARGS, NULL },\n" %
-                     (name, name))
-    elif module == "libvirt-lxc":
-        include.write("libvirt_lxc_%s(PyObject *self, PyObject *args);\n" % (name))
-        export.write("    { (char *)\"%s\", libvirt_lxc_%s, METH_VARARGS, NULL },\n" %
-                     (name, name))
-    elif module == "libvirt-qemu":
-        include.write("libvirt_qemu_%s(PyObject *self, PyObject *args);\n" % (name))
-        export.write("    { (char *)\"%s\", libvirt_qemu_%s, METH_VARARGS, NULL },\n" %
-                     (name, name))
+    include.write("%s_%s(PyObject *self, PyObject *args);\n" % (package, name))
+    export.write("    { (char *)\"%s\", %s_%s, METH_VARARGS, NULL },\n" %
+                 (name, package, name))
 
     if file == "python":
         # Those have been manually generated
@@ -764,22 +767,17 @@ def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[s
             include.write("#endif\n")
             export.write("#endif\n")
             output.write("#endif\n")
-        return 1
+        return True
     if file == "python_accessor" and r_type != "void" and not r_field:
         # Those have been manually generated
         if cond:
             include.write("#endif\n")
             export.write("#endif\n")
             output.write("#endif\n")
-        return 1
+        return True
 
     output.write("PyObject *\n")
-    if module == "libvirt":
-        output.write("libvirt_%s(PyObject *self ATTRIBUTE_UNUSED," % (name))
-    elif module == "libvirt-lxc":
-        output.write("libvirt_lxc_%s(PyObject *self ATTRIBUTE_UNUSED," % (name))
-    elif module == "libvirt-qemu":
-        output.write("libvirt_qemu_%s(PyObject *self ATTRIBUTE_UNUSED," % (name))
+    output.write("%s_%s(PyObject *self ATTRIBUTE_UNUSED," % (package, name))
     output.write(" PyObject *args")
     if format == "":
         output.write(" ATTRIBUTE_UNUSED")
@@ -807,16 +805,7 @@ def print_function_wrapper(module: str, name: str, output: IO[str], export: IO[s
         export.write("#endif /* %s */\n" % cond)
         output.write("#endif /* %s */\n" % cond)
 
-    if module == "libvirt":
-        if name in function_skip_python_impl:
-            return 0
-    elif module == "libvirt-lxc":
-        if name in lxc_function_skip_python_impl:
-            return 0
-    elif module == "libvirt-qemu":
-        if name in qemu_function_skip_python_impl:
-            return 0
-    return 1
+    return True
 
 
 def print_c_pointer(classname: str, output: IO[str], export: IO[str], include: IO[str]) -> None:
@@ -841,23 +830,9 @@ def print_c_pointer(classname: str, output: IO[str], export: IO[str], include: I
                  (classname, classname))
 
 
-def buildStubs(module: str, api_xml: str) -> int:
+def load_apis(module: str, api_xml: str):
     global onlyOverrides
 
-    if module not in ["libvirt", "libvirt-qemu", "libvirt-lxc"]:
-        print("ERROR: Unknown module type: %s" % module)
-        return -1
-
-    if module == "libvirt":
-        funcs = functions
-        funcs_skipped = functions_skipped
-    elif module == "libvirt-lxc":
-        funcs = lxc_functions
-        funcs_skipped = lxc_functions_skipped
-    elif module == "libvirt-qemu":
-        funcs = qemu_functions
-        funcs_skipped = qemu_functions_skipped
-
     try:
         onlyOverrides = False
         with open(api_xml) as stream:
@@ -866,7 +841,7 @@ def buildStubs(module: str, api_xml: str) -> int:
         print(api_xml, ":", msg)
         sys.exit(1)
 
-    n = len(funcs)
+    n = len(functions)
     if not quiet:
         print("Found %d functions in %s" % ((n), api_xml))
 
@@ -883,11 +858,13 @@ def buildStubs(module: str, api_xml: str) -> int:
     if not quiet:
         # XXX: This is not right, same function already in @functions
         # will be overwritten.
-        print("Found %d functions in %s" % (len(funcs) - n, override_api_xml))
+        print("Found %d functions in %s" % (len(functions) - n, override_api_xml))
+
+
+def emit_c_code(module: str) -> None:
+    package = module.replace('-', '_')
+
     nb_wrap = 0
-    failed = 0
-    skipped = 0
-    funcs_failed = []  # type: List[str]
 
     header_file = "build/%s.h" % module
     export_file = "build/%s-export.c" % module
@@ -906,18 +883,8 @@ def buildStubs(module: str, api_xml: str) -> int:
     wrapper.write("#include \"typewrappers.h\"\n")
     wrapper.write("#include \"build/%s.h\"\n\n" % (module,))
 
-    for function in sorted(funcs):
-        # Skip the functions which are not for the module
-        ret = print_function_wrapper(module, function, wrapper, export, include)
-        if ret < 0:
-            failed += 1
-            funcs_failed.append(function)
-            del funcs[function]
-        if ret == 0:
-            skipped += 1
-            funcs_skipped.add(function)
-            del funcs[function]
-        if ret == 1:
+    for function in sorted(functions):
+        if print_function_wrapper(package, function, wrapper, export, include):
             nb_wrap += 1
 
     if module == "libvirt":
@@ -936,20 +903,6 @@ def buildStubs(module: str, api_xml: str) -> int:
     if not quiet:
         print("Generated %d wrapper functions" % nb_wrap)
 
-    if unknown_types:
-        print("Missing type converters: ")
-        for type, count in unknown_types.items():
-            print("%s:%d " % (type, len(count)))
-
-    for f in funcs_failed:
-        print("ERROR: failed %s" % f)
-
-    if failed > 0:
-        return -1
-    if unknown_types:
-        return -1
-    return 0
-
 
 #######################################################################
 #
@@ -1283,24 +1236,27 @@ def functionSortKey(info: Tuple) -> Tuple[str, str]:
 
 
 def writeDoc(module: str, name: str, args: List[ArgumentType], indent: str, output: IO) -> None:
-    if module == "libvirt":
-        funcs = functions
-    elif module == "libvirt-lxc":
-        funcs = lxc_functions
-    elif module == "libvirt-qemu":
-        funcs = qemu_functions
-    if not funcs[name][0]:
+    if not functions[name][0]:
         return
-    val = funcs[name][0]
+    val = functions[name][0]
     val = val.replace("NULL", "None")
     sep = '\n%s' % (indent,)
     output.write('%s"""%s """\n' % (indent, sep.join(val.splitlines())))
 
 
-def buildWrappers(module: str) -> None:
-    if not module == "libvirt":
-        print("ERROR: Unknown module type: %s" % module)
-        return None
+def emit_py_code(module: str) -> None:
+    package = module.replace('-', '_')
+    if module == "libvirt":
+        pymod = "libvirtmod"
+        cygmod = "cygvirtmod"
+    elif module == "libvirt-lxc":
+        pymod = "libvirtmod_lxc"
+        cygmod = "cygvirtmod_lxc"
+    elif module == "libvirt-qemu":
+        pymod = "libvirtmod_qemu"
+        cygmod = "cygvirtmod_qemu"
+    else:
+        raise Exception("Unknown module '%s'" % module)
 
     for tinfo in classes_type.values():
         function_classes[tinfo[2]] = []
@@ -1331,6 +1287,9 @@ def buildWrappers(module: str) -> None:
         ctypes_processed.add(type)
 
     for name, (desc, ret, args, file, mod, cond) in functions.items():
+        if skip_py_impl(name):
+            continue
+
         for type in ctypes:
             classe = classes_type[type][2]
 
@@ -1350,7 +1309,7 @@ def buildWrappers(module: str) -> None:
             info = (0, func, name, ret, args, file, mod)
             function_classes['None'].append(info)
 
-    classes_file = "build/%s.py" % module
+    classes_file = "build/%s.py" % package
     extra_file = "%s-override.py" % module
     extra = None
 
@@ -1369,12 +1328,28 @@ def buildWrappers(module: str) -> None:
     classes.write("#\n")
     classes.write("# WARNING WARNING WARNING WARNING\n")
     classes.write("#\n")
+    classes.write("try:\n")
+    classes.write("    import %s  # type: ignore\n" % pymod)
+    classes.write("except ImportError as lib_e:\n")
+    classes.write("    try:\n")
+    classes.write("        import %s as %s  # type: ignore\n" % (cygmod, pymod))
+    classes.write("    except ImportError as cyg_e:\n")
+    classes.write("        if \"No module named\" in str(cyg_e):\n")
+    classes.write("            raise lib_e\n\n")
+
+    if module != "libvirt":
+        classes.write("import libvirt\n")
+        classes.write("\n")
+
     if extra:
+        classes.write("# WARNING WARNING WARNING WARNING\n")
+        classes.write("#\n")
+        classes.write("# Manually written part of python bindings for %s\n" % module)
         classes.writelines(extra.readlines())
     classes.write("#\n")
     classes.write("# WARNING WARNING WARNING WARNING\n")
     classes.write("#\n")
-    classes.write("# Automatically written part of python bindings for libvirt\n")
+    classes.write("# Automatically written part of python bindings for %s\n" % module)
     classes.write("#\n")
     classes.write("# WARNING WARNING WARNING WARNING\n")
     if extra:
@@ -1414,7 +1389,7 @@ def buildWrappers(module: str) -> None:
                 classes.write("    ret = ")
             else:
                 classes.write("    ")
-            classes.write("libvirtmod.%s(" % name)
+            classes.write("%s.%s(" % (pymod, name))
             for n, (a_name, a_type, a_info) in enumerate(args):
                 if n != 0:
                     classes.write(", ")
@@ -1467,7 +1442,10 @@ def buildWrappers(module: str) -> None:
 
             classes.write("\n")
 
-    for classname in classes_list:
+    modclasses = []
+    if module == "libvirt":
+        modclasses = classes_list
+    for classname in modclasses:
         PARENTS = {
             "virConnect": "self._conn",
             "virDomain": "self._dom",
@@ -1508,8 +1486,8 @@ def buildWrappers(module: str) -> None:
             if classname in classes_destructors:
                 classes.write("    def __del__(self):\n")
                 classes.write("        if self._o is not None:\n")
-                classes.write("            libvirtmod.%s(self._o)\n" %
-                              classes_destructors[classname])
+                classes.write("            %s.%s(self._o)\n" %
+                              (pymod, classes_destructors[classname]))
                 classes.write("        self._o = None\n\n")
                 destruct = classes_destructors[classname]
 
@@ -1528,8 +1506,8 @@ def buildWrappers(module: str) -> None:
 
             classes.write("    def c_pointer(self):\n")
             classes.write("        \"\"\"Get C pointer to underlying object\"\"\"\n")
-            classes.write("        return libvirtmod.%s_pointer(self._o)\n\n" %
-                          classname)
+            classes.write("        return %s.%s_pointer(self._o)\n\n" %
+                          (pymod, classname))
 
             flist = function_classes[classname]
             oldfile = ""
@@ -1574,7 +1552,7 @@ def buildWrappers(module: str) -> None:
                     classes.write("        ret = ")
                 else:
                     classes.write("        ")
-                classes.write("libvirtmod.%s(" % name)
+                classes.write("%s.%s(" % (pymod, name))
                 for n, (a_name, a_type, a_info) in enumerate(args):
                     if n != 0:
                         classes.write(", ")
@@ -1646,8 +1624,9 @@ def buildWrappers(module: str) -> None:
 
                 classes.write("\n")
             # Append "<classname>.py" to class def, iff it exists
-            try:
-                extra = open("libvirt-override-%s.py" % (classname,), "r")
+            class_override = "%s-override-%s.py" % (module, classname)
+            if os.path.exists(class_override):
+                extra = open(class_override, "r")
                 classes.write("    #\n")
                 classes.write("    # %s methods from %s.py (hand coded)\n" % (classname, classname))
                 classes.write("    #\n")
@@ -1660,12 +1639,12 @@ def buildWrappers(module: str) -> None:
 
                 def shouldSkip(lines: List[str]) -> bool:
                     for line in lines:
-                        offset = line.find("libvirtmod.")
+                        offset = line.find(pymod + ".")
                         if offset != -1:
                             func = line[offset + 11:]
                             offset = func.find("(")
                             func = func[0:offset]
-                            if func not in functions_skipped:
+                            if not skip_c_impl(func) and func != "virConnectListDomains":
                                 return True
                     return False
 
@@ -1692,8 +1671,61 @@ def buildWrappers(module: str) -> None:
                     classes.writelines(cached)
                 classes.write("\n")
                 extra.close()
-            except Exception:
-                pass
+
+    direct_functions = {}
+    if module != "libvirt":
+        direct_functions = functions
+
+        classes.write("#\n# Functions from module %s\n#\n\n" % module)
+
+    #
+    # Generate functions directly, no classes
+    #
+    for name, (desc, ret, args, file, mod, cond) in sorted(direct_functions.items()):
+        if skip_py_impl(name):
+            continue
+        func = nameFixup(name, 'None', '', '')
+        classes.write("def %s(" % func)
+        for n, (a_name, a_type, a_info) in enumerate(args):
+            if n != 0:
+                classes.write(", ")
+            classes.write("%s" % a_name)
+        classes.write("):\n")
+        writeDoc(module, name, args, '    ', classes)
+
+        r_type, r_info, r_field = ret
+        if r_type != "void":
+            classes.write("    ret = ")
+        else:
+            classes.write("    ")
+        classes.write("%s.%s(" % (pymod, name))
+
+        conn = None
+
+        for n, (a_name, a_type, a_info) in enumerate(args):
+            if a_type == "virConnectPtr":
+                conn = a_name
+
+            if n != 0:
+                classes.write(", ")
+            if a_type in ["virDomainPtr", "virConnectPtr"]:
+                # FIXME: This might have problem if the function
+                # has multiple args which are objects.
+                classes.write("%s.%s" % (a_name, "_o"))
+            else:
+                classes.write("%s" % a_name)
+        classes.write(")\n")
+
+        if r_type != "void":
+            classes.write("    if ret is None:\n"
+                     "        raise libvirt.libvirtError('%s() failed')\n" % (name,))
+            if r_type == "virDomainPtr":
+                classes.write("    __tmp = libvirt.virDomain(%s, _obj=ret)\n" % (conn,))
+                classes.write("    return __tmp\n")
+            else:
+                classes.write("    return ret\n")
+
+        classes.write("\n")
 
     #
     # Generate enum constants
@@ -1729,238 +1761,33 @@ def buildWrappers(module: str) -> None:
             classes.write("%s = %s\n" % (name, value))
         classes.write("\n")
 
-    classes.write("# typed parameter names\n")
-    for name, value in params:
-        classes.write("%s = \"%s\"\n" % (name, value))
+    if params:
+        classes.write("# typed parameter names\n")
+        for name, value in params:
+            classes.write("%s = \"%s\"\n" % (name, value))
 
     classes.close()
 
+if sys.argv[1] not in ["libvirt", "libvirt-lxc", "libvirt-qemu"]:
+    print("ERROR: unknown module %s" % sys.argv[1])
+    sys.exit(1)
 
-def qemuBuildWrappers(module: str) -> None:
-    if not module == "libvirt-qemu":
-        print("ERROR: only libvirt-qemu is supported")
-        return None
-
-    extra_file = "%s-override.py" % module
-    extra = None
-
-    fd = open("build/libvirt_qemu.py", "w")
-
-    if os.path.exists(extra_file):
-        extra = open(extra_file, "r")
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    fd.write("# This file is automatically written by generator.py. Any changes\n")
-    fd.write("# made here will be lost.\n")
-    fd.write("#\n")
-    fd.write("# To change the manually written methods edit %s-override.py\n" % (module,))
-    fd.write("# To change the automatically written methods edit generator.py\n")
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    fd.write("# Automatically written part of python bindings for libvirt\n")
-    fd.write("#\n")
-
-    fd.write("try:\n")
-    fd.write("    import libvirtmod_qemu\n")
-    fd.write("except ImportError as lib_e:\n")
-    fd.write("    try:\n")
-    fd.write("        import cygvirtmod_qemu as libvirtmod_qemu\n")
-    fd.write("    except ImportError as cyg_e:\n")
-    fd.write("        if \"No module named\" in str(cyg_e):\n")
-    fd.write("            raise lib_e\n\n")
-
-    fd.write("import libvirt\n\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    if extra:
-        fd.writelines(extra.readlines())
-    fd.write("#\n")
-    if extra:
-        extra.close()
-
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    fd.write("#\n# Functions from module %s\n#\n\n" % module)
-    #
-    # Generate functions directly, no classes
-    #
-    for name, (desc, ret, args, file, mod, cond) in sorted(qemu_functions.items()):
-        func = nameFixup(name, 'None', '', '')
-        fd.write("def %s(" % func)
-        for n, (a_name, a_type, a_info) in enumerate(args):
-            if n != 0:
-                fd.write(", ")
-            fd.write("%s" % a_name)
-        fd.write("):\n")
-        writeDoc(module, name, args, '    ', fd)
-
-        r_type, r_info, r_field = ret
-        if r_type != "void":
-            fd.write("    ret = ")
-        else:
-            fd.write("    ")
-        fd.write("libvirtmod_qemu.%s(" % name)
-
-        conn = None
-
-        for n, (a_name, a_type, a_info) in enumerate(args):
-            if a_type == "virConnectPtr":
-                conn = a_name
-
-            if n != 0:
-                fd.write(", ")
-            if a_type in ["virDomainPtr", "virConnectPtr"]:
-                # FIXME: This might have problem if the function
-                # has multiple args which are objects.
-                fd.write("%s.%s" % (a_name, "_o"))
-            else:
-                fd.write("%s" % a_name)
-        fd.write(")\n")
-
-        if r_type != "void":
-            fd.write("    if ret is None:\n"
-                     "        raise libvirt.libvirtError('%s() failed')\n" % (name,))
-            if r_type == "virDomainPtr":
-                fd.write("    __tmp = libvirt.virDomain(%s, _obj=ret)\n" % (conn,))
-                fd.write("    return __tmp\n")
-            else:
-                fd.write("    return ret\n")
-
-        fd.write("\n")
-
-    #
-    # Generate enum constants
-    #
-    for type, enum in sorted(qemu_enums.items()):
-        fd.write("# %s\n" % type)
-        for name, value in sorted(enum.items(), key=lambda i: (int(i[1]), i[0])):
-            fd.write("%s = %s\n" % (name, value))
-        fd.write("\n")
-
-    fd.close()
-
-
-def lxcBuildWrappers(module: str) -> None:
-    if not module == "libvirt-lxc":
-        print("ERROR: only libvirt-lxc is supported")
-        return None
-
-    extra_file = "%s-override.py" % module
-    extra = None
-
-    fd = open("build/libvirt_lxc.py", "w")
-
-    if os.path.exists(extra_file):
-        extra = open(extra_file, "r")
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    fd.write("# This file is automatically written by generator.py. Any changes\n")
-    fd.write("# made here will be lost.\n")
-    fd.write("#\n")
-    fd.write("# To change the manually written methods edit %s-override.py\n" % (module,))
-    fd.write("# To change the automatically written methods edit generator.py\n")
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    if extra:
-        fd.writelines(extra.readlines())
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    fd.write("#\n")
-    fd.write("# Automatically written part of python bindings for libvirt\n")
-    fd.write("#\n")
-    fd.write("# WARNING WARNING WARNING WARNING\n")
-    if extra:
-        extra.close()
-
-    fd.write("try:\n")
-    fd.write("    import libvirtmod_lxc\n")
-    fd.write("except ImportError as lib_e:\n")
-    fd.write("    try:\n")
-    fd.write("        import cygvirtmod_lxc as libvirtmod_lxc\n")
-    fd.write("    except ImportError as cyg_e:\n")
-    fd.write("        if \"No module named\" in str(cyg_e):\n")
-    fd.write("            raise lib_e\n\n")
-
-    fd.write("import libvirt\n\n")
-    fd.write("#\n# Functions from module %s\n#\n\n" % module)
-    #
-    # Generate functions directly, no classes
-    #
-    for name, (desc, ret, args, file, mod, cond) in sorted(lxc_functions.items()):
-        func = nameFixup(name, 'None', '', '')
-        fd.write("def %s(" % func)
-        for n, (a_name, a_type, a_info) in enumerate(args):
-            if n != 0:
-                fd.write(", ")
-            fd.write("%s" % a_name)
-        fd.write("):\n")
-        writeDoc(module, name, args, '    ', fd)
-
-        r_type, r_info, r_field = ret
-        if r_type != "void":
-            fd.write("    ret = ")
-        else:
-            fd.write("    ")
-        fd.write("libvirtmod_lxc.%s(" % name)
-
-        conn = None
-
-        for n, (a_name, a_type, a_info) in enumerate(args):
-            if a_type == "virConnectPtr":
-                conn = a_name
-
-            if n != 0:
-                fd.write(", ")
-            if a_type in ["virDomainPtr", "virConnectPtr"]:
-                # FIXME: This might have problem if the function
-                # has multiple args which are objects.
-                fd.write("%s.%s" % (a_name, "_o"))
-            else:
-                fd.write("%s" % a_name)
-        fd.write(")\n")
-
-        if r_type != "void":
-            fd.write("    if ret is None:\n"
-                     "        raise libvirt.libvirtError('%s() failed')\n" % (name,))
-            if r_type == "virDomainPtr":
-                fd.write("    __tmp = libvirt.virDomain(%s, _obj=ret)\n" % (conn,))
-                fd.write("    return __tmp\n")
-            else:
-                fd.write("    return ret\n")
-
-        fd.write("\n")
-
-    #
-    # Generate enum constants
-    #
-    for type, enum in sorted(lxc_enums.items()):
-        fd.write("# %s\n" % type)
-        for name, value in sorted(enum.items(), key=lambda i: (int(i[1]), i[0])):
-            fd.write("%s = %s\n" % (name, value))
-        fd.write("\n")
-
-    fd.close()
+load_apis(sys.argv[1], sys.argv[2])
 
+if validate_functions() < 0:
+    sys.exit(1)
 
 quiet = False
 if not os.path.exists("build"):
     os.mkdir("build")
 
-if buildStubs(sys.argv[1], sys.argv[2]) < 0:
-    sys.exit(1)
+output = None
+if len(sys.argv) == 4:
+    output = sys.argv[3]
+if output == "c" or output is None:
+    emit_c_code(sys.argv[1])
 
-if sys.argv[1] == "libvirt":
-    buildWrappers(sys.argv[1])
-elif sys.argv[1] == "libvirt-lxc":
-    lxcBuildWrappers(sys.argv[1])
-elif sys.argv[1] == "libvirt-qemu":
-    qemuBuildWrappers(sys.argv[1])
-else:
-    print("ERROR: unknown module %s" % sys.argv[1])
-    sys.exit(1)
+if output == "py" or output is None:
+    emit_py_code(sys.argv[1])
 
 sys.exit(0)
